﻿using JudgeModule;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.UI;

public class JudgeManager : MonoBehaviour
{
    float latency = 225f;
    float scrollMultiplier = 1.0f;
    float startOffset = -5f;
    float baseTime;
    bool IsInitTime;

    bool IsGameEnd;

    private GameObject offset;
    private GameObject noteobj;
    private MotionGageManager motionGageManager;
    private GameObject judgeText;

    private GameObject motionSampleDisplayPrefab;

    private Vector3 initialPos;

    private float MsPerBeat
    {
        get
        {
            return 60 * 1000f / GameManager.Instance.CurrentTrack.BPM;
        }
    }
    
    private float ScrollSpeed
    {
        get
        {
            return scrollMultiplier * MsPerBeat / 1000f;
        }
    }

    private InputManager input;
    private NoteCondition condition;
    private List<Note> notes;

    void SetObjectRef()
    {
        offset = GameObject.Find("Offset");
        noteobj = GameObject.Find("Noteobj");
        judgeText = GameObject.Find("Judge");
        input = GameObject.Find("InGameManagers").GetComponent<InputManager>();
        condition = new NoteCondition(input);
        motionGageManager = new MotionGageManager
        { motionGuage = GameObject.Find("Motion Guage") };
    }

    void JudgeInit()
    {
        initialPos = offset.transform.position;
        judgeText.SetActive(false);
        motionGageManager.ResetGuage();
        IsGameEnd = false;
        baseTime = 0f;
        IsInitTime = false;
    }

    void LoadGameObjects()
    {
        motionSampleDisplayPrefab = Resources.Load("Motion Sample Display") as GameObject;
        notes = GameManager.Instance.CurrentTrack.Notes.ToList();
        notes.ForEach(LoadNote);

        var sounds = GameObject.Find("Sounds");
        sounds.transform.Cast<Transform>()
                        .ToList()
                        .ForEach(x => x.gameObject.SetActive(false));
        if (GameManager.Instance.CurrentTrack.BGM != null)
            sounds.transform.Find(GameManager.Instance.CurrentTrack.BGM)
                  .gameObject.SetActive(true);
    }
    
    // Use this for initialization
    void Start()
    {
        SetObjectRef();
        JudgeInit();
        LoadGameObjects();
    }

    // Update is called once per frame
    void Update()
    {
        float current = motionGageManager.UpdateTime();
        if (!IsInitTime)
        {
            baseTime = current;
            IsInitTime = true;
        }

        float timing = current - baseTime;

        offset.transform.position = Vector3.left * current * ScrollSpeed;
        
        motionGageManager.UpdateGuage();

        if (noteobj.transform.childCount == 0 ||
            IsGameEnd)
            Invoke("ShowResult", 2f);

        JudgeNote(timing);
    }

    void JudgeNote(float timing)
    {
        Note note = GetLastNote(timing);
        if (note == null)
            return;

        if (condition.IsWrongInput(note))
        {
            WrongNoteProc(note);
            return;
        }

        if (Judge.IsNoteStart(note, timing))
            PlaySampleMotion(note as MotionNote);

        if (note.IsLong)
            JudgeLongNote(note, timing);
        else
            JudgeShortNote(note, timing);
    }

    Note GetLastNote(float timing)
    {
        while (noteobj.transform.childCount != 0)
        {
            var note = noteobj.transform.GetChild(0).gameObject
                              .GetComponent<Note.Controller>().Instance;
            if (!Judge.IsPastNote(note, timing))
                return note;
            SetJudge(Judge.MISS);
            DeactivateNote(note);
        }

        return null;
    }

    void PlaySampleMotion(MotionNote note)
    {
        GameObject motionSample = Instantiate(motionSampleDisplayPrefab);
        MotionSampleDisplay msd = motionSample.GetComponent<MotionSampleDisplay>();
        msd.sprite = note.Image;
        msd.timeout = MsPerBeat;
        note.Activated = true;
        note.MotionSampleDisplay = msd;
    }

    void JudgeShortNote(Note note, float timing)
    {
        if (condition.IsShortNoteEntered(note))
            EnteredNoteProc(note, timing);
    }

    void JudgeLongNote(Note note, float timing)
    {
        if (condition.IsLongNoteStartCorrectly(note))
            EnteredNoteProc(note, timing);
        if (condition.IsLongNoteHoldCorrectly(note))
            ContinuingNoteProc(note, timing);
        if (condition.IsLongNoteFinishCorrectly(note, timing))
            CorrectlyStoppedNoteProc(note, timing);
        if (condition.IsLongNoteFinishIncorrectly(note, timing))
            IncorrectlyStoppedNoteProc(note, timing);
    }

    void WrongNoteProc(Note note)
    {
        SetJudge(Judge.MISS);
        DeactivateNote(note);

        if (!(note is MotionNote))
        {
            IsGameEnd = true;
            Invoke("ShowResult", 2f);
        }
    }

    void EnteredNoteProc(Note note, float timing)
    {
        SetJudge(Judge.TestJudge(note, timing));

        if (note.IsLong)
            note.Activated = true;
        else
            DeactivateNote(note);
    }

    void ContinuingNoteProc(Note note, float timing)
    {
        bool isNoteEnd = Judge.IsNoteEnd(note, timing);
        var judge = Judge.TestJudge(note, timing, isNoteEnd);
        SetJudge(judge);

        if (isNoteEnd ||
            condition.IsNoteBroken(note, timing, judge))
            DeactivateNote(note);
    }

    void CorrectlyStoppedNoteProc(Note note, float timing)
    {
        SetJudge(Judge.TestJudge(note, timing, true));
        DeactivateNote(note);
    }

    void IncorrectlyStoppedNoteProc(Note note, float timing)
    {
        var judge = Judge.TestJudge(note, timing, note.Activated);
        SetJudge(Judge.IsNoteComboBroken(judge) ? judge : Judge.MISS);
        DeactivateNote(note);
    }

    private void DeactivateNote(Note note)
    {
        note.Activated = false;
        note.Component.transform.SetParent(offset.transform);
        note.Component.Deactivate();
    }

    private void SetJudge(Judge judge, bool isMotion = false)
    {
        if (!judge.IsBreak)
            ++GameManager.Instance.Combo;
        else
            GameManager.Instance.Combo = 0;
        GameManager.Instance.Score += judge.Score;
        GameManager.Instance.JudgeCount[judge]++;

        Debug.Log(judge.Name + (isMotion ? " Motion" : " Note") + " Combo: " + GameManager.Instance.Combo);

        judgeText.SetActive(true);
        judgeText.GetComponent<Text>().text = judge.Name;
        judgeText.GetComponent<Text>().color = judge.Color;
    }

    void LoadNote(Note note)
    {
        GameObject obj = Instantiate(
                Resources.Load(note.Type.ToString(), typeof(GameObject)),
                noteobj.transform)
                as GameObject;

        obj.AddComponent<Note.Controller>().Instance = note;

        if (note.Type == NoteType.MeasureLine || note.Type == NoteType.BeatLine)
            obj.transform.SetParent(offset.transform);

        if (note.IsLong)
            StretchLongNote(note, obj);

        obj.transform.position += initialPos +
            (Vector3.right * (note.StartTiming * ScrollSpeed));
    }

    void StretchLongNote(Note note, GameObject obj)
    {
        float length = note.Length * ScrollSpeed;

        var holdTransform = obj.transform.GetChild(1)
            .GetComponent<RectTransform>();
        var endTransform = obj.transform.GetChild(2)
            .GetComponent<RectTransform>();

        holdTransform.SetSizeWithCurrentAnchors(
            RectTransform.Axis.Horizontal, length);
        holdTransform.position += (Vector3.right * length / 2);
        endTransform.position += (Vector3.right * length);
    }

    private bool onResult = false;

    private void ShowResult()
    {
        if (!onResult)
        {
            Destroy(GetComponent<InputManager>());
            Destroy(GetComponent<MotionView>());
            Destroy(GetComponent<CoordinateMapperManager>());
            Destroy(this);

            GameManager.Instance.SceneTransition("Result");
            onResult = true;
        }
    }
}
